﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.IO.Ports;
using NDesk.Options;
using System.Text.RegularExpressions;
using System.Threading;

namespace SerialPlayer
{
    class Program
    {

        static bool show_help = false;
        static bool show_ports = false;
        
        static string filename = null;
        static string fileformat = null;

        static Dictionary<UartFileReader.AttributeType, object> cmdparams = new Dictionary<UartFileReader.AttributeType, object>();

        static bool ParseUartCfg(string cfg)
        {
            Regex reConfig = new Regex(@"^([5-8])([NOEMS])(0|1\.5|1,5|1|2)$", RegexOptions.IgnoreCase);
            Match m = reConfig.Match(cfg);
            if (!m.Success)
                return false;
            
            // bits
            cmdparams.Add(UartFileReader.AttributeType.UartDataBits, int.Parse(m.Groups[1].Value));
            bool success = true;

            // parity
            Parity? parity = null;
            switch (m.Groups[2].Value)
            {
                case "N": parity = Parity.None; break;
                case "O": parity = Parity.Odd; break;
                case "E": parity = Parity.Even; break;
                case "M": parity = Parity.Mark; break;
                case "S": parity = Parity.Space; break;
                default: success = false; break;
            }
            if(parity != null)
            {
                cmdparams.Add(UartFileReader.AttributeType.UartParity, (Parity)parity);
            }

            // stopbits
            StopBits? stopbits = null;
            switch(m.Groups[3].Value)
            {
                case "0": stopbits = StopBits.None; break;
                case "1": stopbits = StopBits.One; break;
                case "1.5":
                case "1,5":
                    stopbits = StopBits.OnePointFive; break;
                case "2":
                    stopbits = StopBits.Two; break;
                default: success = false; break;
            }
            if(stopbits != null)
            {
                cmdparams.Add(UartFileReader.AttributeType.UartStopBits, (StopBits)stopbits);
            }

            return success;
        }

        enum SettingSource
        {
            None = 0,
            Parameter = 1,
            File = 2,
            Both = 3,
        }

        static SettingSource GetSetting(Dictionary<UartFileReader.AttributeType, object> param, Dictionary<UartFileReader.AttributeType, object> file, UartFileReader.AttributeType attribute, out object value)
        {
            int foo = 0;
            if (param.ContainsKey(attribute))
                foo |= 1;
            if (file.ContainsKey(attribute))
                foo |= 2;

            SettingSource result = (SettingSource)foo;

            switch (result)
            {
                case SettingSource.None:
                    Console.WriteLine("Parameter '{0}' not set.", attribute);
                    value = null; 
                    break;
                case SettingSource.Parameter: 
                    value = param[attribute]; 
                    break;
                case SettingSource.File: 
                    value = file[attribute];  
                    break;
                case SettingSource.Both:
                    Console.WriteLine("Parameter {0} from file overridden by argument.", attribute);
                    value = param[attribute]; 
                    break;
                default: 
                    value = null; 
                    break;
            }

            return result;
        }

        static void Main(string[] args)
        {

            OptionSet options = new OptionSet()
            {
                { "h|help", "Show this help and exit", v => show_help = v != null },
                { "l|list", "List available COM-ports", v => show_ports = v != null },
                { "p|port=", "The serial port you want the data to output (COMx)", 
                    v => cmdparams.Add(UartFileReader.AttributeType.Portname, v) },
                { "b|baudrate=", "Baudrate to be used for communication", 
                    (int v) => cmdparams.Add(UartFileReader.AttributeType.Baudrate, v) },
                { "c|config=", "Configuration for the communication", 
                    v => ParseUartCfg(v) },
                { "fn|file=", "File used for transmission", v => filename = v },
                { "ff|fileformat=", "The file's format: tsv-asc, tsv-hex (default: tsv-asc)", v => fileformat = v },
                { "rc|repeat-count=", "Repetition count of output",
                    (int v) => cmdparams.Add(UartFileReader.AttributeType.RepetitionCount, v) },
                { "rp|repeat-pause=", "Pause between repetitions in ms",
                    (int v) => cmdparams.Add(UartFileReader.AttributeType.RepetitionPause, v) },
            };

            string exename = AppDomain.CurrentDomain.FriendlyName;
            List<string> extra = null;
            try
            {
                extra = options.Parse(args);
            }
            catch(OptionException ex)
            {
                Console.WriteLine("Error: {0}", ex.Message);
                Console.WriteLine("Try '{0} --help' for more information", exename);
                return;
            }

            if(show_help)
            {
                Console.WriteLine("Replays recorded files to a Serial port");
                Console.WriteLine();
                Console.WriteLine("{0} -fn filename [-ff tsv-asc|tsv-hex] [-p port] [-b baudrate] [-c config] [--rc repetitions] [--rp pause]", exename);
                options.WriteOptionDescriptions(Console.Out);

                Console.WriteLine();
                Console.WriteLine("Examples:");
                Console.WriteLine(" {0} -f recording.txt -p COM1 -b 115200 -c 8N1");
                Console.WriteLine(); 
                Console.WriteLine("For tsv-asc, data files must be in the following format:");
                Console.WriteLine("<timestamp>\\t<data>\\n");
                Console.WriteLine(" timestamp: absolute time in seconds, milliseconds separated by '.'");
                Console.WriteLine(" data: data as ASCII string, escape sequences like \\0, \\a or \\x42 are allowed, no unicode.");
                Console.WriteLine("");
                Console.WriteLine("Lines starting with '#' will be ignored, if the first line starts with #OPT, the contents will be interpreted with the following Regex:");
                Console.WriteLine(@" /^#\s*OPT\s+(<baudrate>\d+)\s+(<databits>[5-8])(<parity>[NOEMS])(<stopbits>0|1\.5|1,5|1|2)/i");

                return;
            }

            if(show_ports)
            {
                Console.WriteLine("Available Ports:");
                foreach (string p in SerialPort.GetPortNames())
                {
                    Console.WriteLine("- {0}", p);
                }
                return;
            }

            if (filename == null)
            {
                Console.WriteLine("No input file given.");
                return;
            } else if (!System.IO.File.Exists(filename))
            {
                Console.WriteLine("File '{0}' does not exist.", filename);
                return;
            }

            StreamReader sr = null;
            try
            {
                sr = new StreamReader(filename);
            }
            catch (Exception e)
            {
                Console.WriteLine("Could not open file: '{0}'.", e.Message);
                return;
            }

            Console.WriteLine("Loading '{0}' as tsv-ascii...", filename);
            UartFileReader reader = null;
            try
            {
                switch (fileformat)
                {
                    case null:
                    case "tsv-asc":
                        reader = new UfrTsvAsc(sr);
                        break;
                    case "tsv-hex":
                        reader = new UfrTsvHex(sr);
                        break;
                    default:
                        Console.WriteLine("Unknown file format '{0}'", fileformat);
                        Console.WriteLine("Aborted");
                        return;
                }
                
            }
            catch(Exception ex)
            {
                Console.WriteLine("Could not load file: {0}", ex.Message);
                return;
            }
            
            Console.WriteLine("{0} Packets, {1} Bytes, {2} seconds", reader.data.Count, reader.TotalBytes, reader.Duration);

            // re-validate UART parameters
            string portname = "";
            int baudRate = 0;
            int dataBits = 8;
            Parity parity = Parity.None;
            StopBits stopBits = StopBits.One;

            int repeatcount = 1;
            int repeatpause = 0;

            bool error = false;
            object tmp;

            if (GetSetting(cmdparams, reader.Attributes, UartFileReader.AttributeType.Portname, out tmp) == SettingSource.None)
                error = true;
            else
            {
                portname = (string)tmp;

                if (portname == null)
                {
                    error = true;
                    Console.WriteLine("No port set");
                }
                else if(!SerialPort.GetPortNames().Contains(portname))
                {
                    error = true; 
                    Console.WriteLine("Port '{0}' is not available.", portname);
                }
            }

            if (GetSetting(cmdparams, reader.Attributes, UartFileReader.AttributeType.Baudrate, out tmp) == SettingSource.None)
                error = true;
            else
            {
                baudRate = (int)tmp;
                if(baudRate <= 0)
                {
                    error = true;
                    Console.WriteLine("Invalid baudrate ({0})", baudRate);
                }
            }

            if (GetSetting(cmdparams, reader.Attributes, UartFileReader.AttributeType.UartDataBits, out tmp) == SettingSource.None)
                error = true;
            else
                dataBits = (int)tmp; // no checks needed as it is already done

            if (GetSetting(cmdparams, reader.Attributes, UartFileReader.AttributeType.UartParity, out tmp) == SettingSource.None)
                error = true;
            else
                parity = (Parity)tmp; // no checks needed as it is already done

            if (GetSetting(cmdparams, reader.Attributes, UartFileReader.AttributeType.UartStopBits, out tmp) == SettingSource.None)
                error = true;
            else
                stopBits = (StopBits)tmp; // no checks needed as it is already done

            if (GetSetting(cmdparams, reader.Attributes, UartFileReader.AttributeType.RepetitionCount, out tmp) != SettingSource.None)
            {
                repeatcount = (int)tmp;
                if(repeatcount <= 0)
                {
                    error = true;
                    Console.WriteLine("Invalid repetition count ({0})", repeatcount);
                }
                else
                {
                    if (GetSetting(cmdparams, reader.Attributes, UartFileReader.AttributeType.RepetitionPause, out tmp) != SettingSource.None)
                    {
                        repeatpause = (int)tmp;
                        if(repeatpause < 0)
                        {
                            error = true;
                            Console.WriteLine("Invalid repetition pause ({0} ms)", repeatpause);
                        }
                    }
                }
            }

            if (error == true)
            {
                Console.WriteLine("Could not continue due to errors.");
                return;
            }

            SerialPort port = null;
            
            try
            {
                Console.WriteLine("Opening '{0}' with {1} Baud, {2} data bits, {3} parity, {4} stop bit(s)", portname, baudRate, dataBits, parity, stopBits);
                port = new SerialPort(portname, baudRate, parity, dataBits, stopBits);
                port.Open();
            }
            catch(Exception e)
            {
                Console.WriteLine("Could not open serial port: {0}", e.Message);
                return;
            }

            int curtoprpt = Console.CursorTop;
            int curtopprog = curtoprpt;
            

            int conwid = Console.BufferWidth;

            for(int c = 0; c < repeatcount; c++)
            {
                if(repeatcount > 1)
                {
                    Console.CursorTop = curtoprpt;
                    Console.WriteLine("Repetition {0} of {1}", c+1, repeatcount);
                    curtopprog = curtoprpt + 1;
                }

                int lastProgressInt = -1;
                double lastTime = 0;
                // sleep time between packets is determined by the actual time to avoid drifts
                DateTime startDt = DateTime.Now;
                for (int i = 0; i < reader.data.Count; i++)
                {
                    DataEntry de = reader.data[i];

                    double sleeptime = (de.timeOffset - lastTime) - (DateTime.Now - startDt).TotalSeconds;

                    if (sleeptime > 0)
                    {
                        Thread.Sleep(new TimeSpan((long)(sleeptime * 10000000)));
                    }

                    try
                    {
                        port.Write(de.data, 0, de.data.Length);
                        port.ReadExisting(); // dummy read
                    }
                    catch(Exception ex)
                    {
                        Console.WriteLine("Could not write to port: {0}", ex.Message);
                        Console.WriteLine("Aborting.");
                        return;
                    }

                    if (reader.Duration > 0)
                    {
                        double progress = progress = de.timeOffset / reader.Duration;
                        if (progress != double.NaN)
                        {
                            int progressInt = (int)(progress * 100);
                            if (progress > 100) progress = 100;

                            if (progressInt != lastProgressInt)
                            {
                                //only print to console when necessary
                                Console.CursorTop = curtopprog;
                                Console.CursorLeft = 0;
                                Console.Write("[");
                                Console.Write("".PadRight(progressInt / 2, '#').PadRight(50, ' '));
                                Console.WriteLine("] {0,3} % ", progressInt);
                            }

                            lastProgressInt = progressInt;
                        }
                    }
                }

                if(repeatpause > 0 && c < (repeatcount - 1))
                {
                    Thread.Sleep(repeatpause);
                }
            }

            port.Close();

            Console.WriteLine("done.");
        }
    }
}
