﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;
using System.IO.Ports;
using System.Globalization;

namespace SerialPlayer
{
    public class UfrTsvHex : UartFileReader
    {

        StreamReader reader = null;
        long lineno = 0;
        string line = "";

        public UfrTsvHex(StreamReader reader)
        {
            this.reader = reader;

            if (!readHeader()) return;
            readFile();
        }

        bool readNextLine()
        {
            if (reader.EndOfStream) return false;
            line = reader.ReadLine();
            lineno++;

            return true;
        }

        bool ParseUartCfg(string cfg)
        {
            Regex reConfig = new Regex(@"^([5-8])([NOEMS])(0|1\.5|1,5|1|2)$", RegexOptions.IgnoreCase);
            Match m = reConfig.Match(cfg);
            if (!m.Success)
                return false;

            // bits, no need to tryparse, the regex did handle that
            int databits = int.Parse(m.Groups[1].Value);
            bool success = true;

            // parity
            Parity parity = Parity.None;
            switch (m.Groups[2].Value)
            {
                case "N": parity = Parity.None; break;
                case "O": parity = Parity.Odd; break;
                case "E": parity = Parity.Even; break;
                case "M": parity = Parity.Mark; break;
                case "S": parity = Parity.Space; break;
                default: success = false; break;
            }

            // stopbits
            StopBits stopbits = StopBits.One;
            switch (m.Groups[3].Value)
            {
                case "0": stopbits = StopBits.None; break;
                case "1": stopbits = StopBits.One; break;
                case "1.5":
                case "1,5":
                    stopbits = StopBits.OnePointFive; break;
                case "2":
                    stopbits = StopBits.Two; break;
                default: success = false; break;
            }

            if(success == true)
            {
                Attributes.Add(AttributeType.UartDataBits, databits);
                Attributes.Add(AttributeType.UartParity, parity);
                Attributes.Add(AttributeType.UartStopBits, stopbits);
            }
            return success;
        }

        bool readHeader()
        {
            if (!readNextLine()) 
                return false;

            Match mOptions = new Regex(@"^#\s*OPT\s+(\d+)\s+([5-8][NOEMS](?:0|1\.5|1,5|1|2))$", RegexOptions.IgnoreCase).Match(line);
            if (mOptions.Success == true)
            {
                int baudrate = int.Parse(mOptions.Groups[1].Value);
                
                if (baudrate <= 0)
                {
                    Console.WriteLine("Invalid baudrate, ignored.");
                }
                else
                {
                    Attributes.Add(AttributeType.Baudrate, baudrate);
                }

                if (!ParseUartCfg(mOptions.Groups[2].Value) == false)
                {
                    //Console.WriteLine("Some port parameters might not be set due to format errors.");
                }

                if (!readNextLine()) 
                    return false;
            }

            return true;
        }

        //https://stackoverflow.com/a/311179
        public static byte[] HexstringToBytearray(String hex)
        {
            int NumberChars = hex.Length;
            byte[] bytes = new byte[NumberChars / 2];
            for (int i = 0; i < NumberChars; i += 2)
                bytes[i / 2] = Convert.ToByte(hex.Substring(i, 2), 16);
            return bytes;
        }

        void readFile()
        {
            TotalBytes = 0;
            Duration = 0;
            double timeOffsetStart = double.NaN;
            while (true)
            {
                if (line.StartsWith("#"))
                {
                    // ignore commented line
                    if (!readNextLine()) return;
                    continue;
                }

                string[] parts = line.Split('\t');
                if (parts.Length != 2)
                {
                    Console.WriteLine("Invalid format in line {0}, ignored.", lineno);
                }

                double timeOffset;
                if (double.TryParse(parts[0], NumberStyles.Any, CultureInfo.InvariantCulture, out timeOffset) == false)
                {
                    Console.WriteLine("Invalid timestamp in line {0}, ignored.", lineno);
                    if (!readNextLine()) return;
                    continue;
                }

                if (double.IsNaN(timeOffsetStart))
                {
                    Attributes.Add(AttributeType.StartOffset, timeOffset);
                    timeOffsetStart = timeOffset;
                }

                byte[] bytes = null;
                try
                {
                    bytes = HexstringToBytearray(parts[1]);
                } 
                catch(Exception ex)
                {
                    Console.WriteLine("Could not read line {0}: {1}", lineno, ex.Message);
                    if (!readNextLine()) return;
                    continue;
                }

                TotalBytes += bytes.Length;

                data.Add(new DataEntry(timeOffset - timeOffsetStart, bytes));

                //Console.WriteLine(">> {0}", outdata);

                if (!readNextLine()) break;
            }

            if(data.Count > 0)
            {
                Duration = data.Last().timeOffset;
            }
        }

    }
}
